@@ -80,6 +80,7 @@ class lua:
 		lua._lib.lua_touserdata.restype = ctypes.c_void_p
 		lua._lib.lua_tothread.restype = ctypes.c_void_p
 		lua._lib.lua_tocfunction.restype = ctypes.c_void_p
+		lua._lib.lua_newuserdata.restype = ctypes.c_void_p
 		self._factory = ctypes.CFUNCTYPE (ctypes.c_int, ctypes.c_void_p)
 		self._objects = {}
 		self._the_add = self._factory (_object_add)
@@ -133,7 +134,7 @@ class lua:
 		lua._lib.lua_setfield (self._state, -2, '__gc')
 		lua._lib.lua_setfield (self._state, LUA_REGISTRYINDEX, 'metatable')
 		if not debug:
-			self.run ('debug = nil')
+			self.run ('debug = nil', name = 'disabling debug')
 	def __del__ (self):
 		del lua._objects[self._state]
 	def run_file (self, script):
@@ -185,14 +186,22 @@ class lua:
 					k = key
 				module[k] = getattr (value, key)
 			value = module
-		self.run ('module ("%s")' % name, name, value)
+		# This must be a real lua table, so don't use _push.
+		lua._lib.lua_createtable (self._state, 0, 0)
+		pos = lua._lib.lua_gettop (self._state)
+		for i in value:
+			self._push (i)
+			self._push (value[i])
+			lua._lib.lua_settable (self._state, pos)
+		lua._lib.lua_setfield (self._state, LUA_GLOBALSINDEX, name)
+		self.run ('module ("%s")' % name, name = 'loading module %s' % name)
 	def _to_python (self, index):
 		_dprint ('creating python value from lua at %d' % index, 1)
 		type = lua._lib.lua_type (self._state, index)
 		if type == LUA_TNIL:
 			return None
 		elif type == LUA_TBOOLEAN:
-			return lua._lib.lua_toboolean (self._state, index)
+			return bool (lua._lib.lua_toboolean (self._state, index))
 		elif type == LUA_TLIGHTUSERDATA:
 			return lua._lib.lua_touserdata (self._state, index)
 		elif type == LUA_TNUMBER:
@@ -200,14 +209,9 @@ class lua:
 		elif type == LUA_TSTRING:
 			return lua._lib.lua_tolstring (self._state, index, None)
 		elif type == LUA_TTABLE:
-			ret = {}
-			lua._lib.lua_pushnil (self._state)
-			if index < 0:
-				index -= 1
-			while lua._lib.lua_next (self._state, index) != 0:
-				ret[self._to_python (-2)] = self._to_python (-1)
-				lua._lib.lua_settop (self._state, -2)
-			return ret
+			_dprint ('creating table', 1)
+			lua._lib.lua_pushvalue (self._state, index)
+			return Table (self)
 		elif type == LUA_TFUNCTION:
 			lua._lib.lua_pushvalue (self._state, index)
 			return Function (self)
@@ -221,7 +225,7 @@ class lua:
 		else:
 			raise AssertionError ('unexpected lua type')
 	def _push (self, obj):
-		_dprint ('pushing %s to lua stack' % str (obj), 1)
+		_dprint ('pushing %s (%s) to lua stack' % (str (obj), type (obj)), 1)
 		if obj is None:
 			lua._lib.lua_pushnil (self._state)
 		elif isinstance (obj, bool):
@@ -232,22 +236,9 @@ class lua:
 			lua._lib.lua_pushlstring (self._state, obj, len (obj))
 		elif isinstance (obj, float):
 			lua._lib.lua_pushnumber (self._state, ctypes.c_double (obj))
-		elif isinstance (obj, Function):
+		elif isinstance (obj, (Function, Table)):
+			_dprint ('pushing table', 1)
 			lua._lib.lua_rawgeti (self._state, LUA_REGISTRYINDEX, obj._id)
-		elif isinstance (obj, list) or isinstance (obj, tuple):
-			lua._lib.lua_createtable (self._state, len (obj), 0)
-			pos = lua._lib.lua_gettop (self._state)
-			for i in range (len (obj)):
-				lua._lib.lua_pushinteger (self._state, i)
-				self._push (obj[i])
-				lua._lib.lua_settable (self._state, pos)
-		elif isinstance (obj, dict):
-			lua._lib.lua_createtable (self._state, 0, 0)
-			pos = lua._lib.lua_gettop (self._state)
-			for i in obj:
-				self._push (i)
-				self._push (obj[i])
-				lua._lib.lua_settable (self._state, pos)
 		elif isinstance (obj, object):
 			o = {}
 			for key in dir (obj):
@@ -266,6 +257,7 @@ class lua:
 		else:
 			# This shouldn't be possible: everything is an object.
 			raise ValueError ('object of type %s cannot be converted to lua object' % str (type (obj)))
+		_dprint ('done pushing', 1)
 
 def _object_add (state):
 	_dprint ('adding lua stuff', 1)
@@ -474,7 +466,10 @@ def _object_put (state):
 				setattr (obj, '_lua' + key[1:], value)
 				_dprint ('set %s[%s] = %s' % (str (obj), str (key), str (value)), 3)
 				return 0
-		setattr (obj, key, value)
+		if isinstance (obj, dict):
+			obj[key] = value
+		else:
+			setattr (obj, key, value)
 		_dprint ('set %s[%s] = %s' % (str (obj), str (key), str (value)), 3)
 		return 0
 	except:
@@ -543,13 +538,93 @@ class Function:
 		_dprint ('destroying lua function', 1)
 		lua._lib.luaL_unref (self._luaobject._state, LUA_REGISTRYINDEX, self._id)
 
-def as_list (table):
-	'Get the values with numerical keys close to integers into a list'
-	_dprint ('get list from lua table', 1)
-	epsilon = 1e-10
-	l = [(int (x + .5), table[x]) for x in table if isinstance (x, numbers.Number) and -epsilon < x - int (x) < epsilon]
-	l.sort (key = lambda x:x[0])
-	ret = [None] * l[-1][0]
-	for i in l:
-		ret[i[0] - 1] = i[1]
-	return ret
+class Table:
+	def __init__ (self, luaobject):
+		_dprint ('creating lua table', 1)
+		self._luaobject = luaobject
+		self._id = lua._lib.luaL_ref (self._luaobject._state, LUA_REGISTRYINDEX)
+	def __len__ (self):
+		_dprint ('requesting length of lua table', 3)
+		lua._lib.lua_rawgeti (self._luaobject._state, LUA_REGISTRYINDEX, self._id)
+		lua._lib.lua_objlen (self._luaobject._state, -1)
+		ret = self._luaobject._to_python (-1)
+		lua._lib.lua_settop (self._luaobject._state, -3)
+		return ret
+	def __getitem__ (self, key):
+		_dprint ('requesting item of lua table: %s' % key, 3)
+		lua._lib.lua_rawgeti (self._luaobject._state, LUA_REGISTRYINDEX, self._id)
+		self._luaobject._push (key)
+		lua._lib.lua_gettable (self._luaobject._state, -2)
+		ret = self._luaobject._to_python (-1)
+		lua._lib.lua_settop (self._luaobject._state, -3)
+		return ret
+	def __setitem__ (self, key, value):
+		_dprint ('setting item of lua table: %s: %s' % (key, value), 3)
+		lua._lib.lua_rawgeti (self._luaobject._state, LUA_REGISTRYINDEX, self._id)
+		self._luaobject._push (key)
+		self._luaobject._push (value)
+		lua._lib.lua_settable (self._luaobject._state, -3)
+		lua._lib.lua_settop (self._luaobject._state, -2)
+	def __contains__ (self, key):
+		_dprint ('checking if %s is in lua table' % key, 3)
+		lua._lib.lua_rawgeti (self._luaobject._state, LUA_REGISTRYINDEX, self._id)
+		lua._lib.lua_pushnil (self._luaobject._state)
+		while lua._lib.lua_next (self._luaobject._state, -2) != 0:
+			if self._luaobject._to_python (-2) == key:
+				lua._lib.lua_settop (self._luaobject._state, -4)
+				return True
+			lua._lib.lua_settop (self._luaobject._state, -2)
+		lua._lib.lua_settop (self._luaobject._state, -2)
+		return False
+	def __iter__ (self):
+		_dprint ('iterating over lua table', 3)
+		lua._lib.lua_rawgeti (self._luaobject._state, LUA_REGISTRYINDEX, self._id)
+		lua._lib.lua_pushnil (self._luaobject._state)
+		while lua._lib.lua_next (self._luaobject._state, -2) != 0:
+			ret = self._luaobject._to_python (-2)
+			lua._lib.lua_settop (self._luaobject._state, -2)
+			try:
+				ref = lua._lib.luaL_ref (self._luaobject._state, LUA_REGISTRYINDEX)
+				lua._lib.lua_settop (self._luaobject._state, -2)
+				yield ret
+				lua._lib.lua_rawgeti (self._luaobject._state, LUA_REGISTRYINDEX, self._id)
+				lua._lib.lua_rawgeti (self._luaobject._state, LUA_REGISTRYINDEX, ref)
+			finally:
+				lua._lib.luaL_unref (self._luaobject._state, LUA_REGISTRYINDEX, ref)
+		lua._lib.lua_settop (self._luaobject._state, -2)
+	def __del__ (self):
+		_dprint ('destroying lua table', 1)
+		lua._lib.luaL_unref (self._luaobject._state, LUA_REGISTRYINDEX, self._id)
+	def __eq__ (self, other):
+		if not isinstance (other, Table):
+			return False
+		lua._lib.lua_rawgeti (self._luaobject._state, LUA_REGISTRYINDEX, self._id)
+		lua._lib.lua_rawgeti (self._luaobject._state, LUA_REGISTRYINDEX, other._id)
+		ret = bool (lua._lib.lua_rawequal (self._luaobject._state, -1, -2))
+		lua._lib.lua_settop (self._luaobject._state, -3)
+		return ret
+	def __ne__ (self, other):
+		return not self == other
+	def _dict (self):
+		'Get a copy of the table as a dict'
+		_dprint ('get dict from lua table', 1)
+		ret = {}
+		lua._lib.lua_rawgeti (self._luaobject._state, LUA_REGISTRYINDEX, self._id)
+		lua._lib.lua_pushnil (self._luaobject._state)
+		while lua._lib.lua_next (self._luaobject._state, -2) != 0:
+			ret[self._luaobject._to_python (-2)] = self._luaobject._to_python (-1)
+			lua._lib.lua_settop (self._luaobject._state, -2)
+		lua._lib.lua_settop (self._luaobject._state, -2)
+		return ret
+	def _list (self):
+		'Get a copy of the table, which must be a sequence, as a list'
+		_dprint ('get list from lua table', 1)
+		lua._lib.lua_rawgeti (self._luaobject._state, LUA_REGISTRYINDEX, self._id)
+		l = lua._lib.lua_objlen (self._luaobject._state, -1)
+		ret = [None] * l
+		for i in range (l):
+			lua._lib.lua_rawgeti (self._luaobject._state, -1, i + 1)
+			ret[i] = self._luaobject._to_python (-1)
+			lua._lib.lua_settop (self._luaobject._state, -2)
+		lua._lib.lua_settop (self._luaobject._state, -2)
+		return ret
